from PCIDev import PCINetDevices
import os
import json
from charmhelpers.core.services import helpers
from charmhelpers.core.hookenv import(
    config,
    log,
)

VLAN = 'vlan'
VXLAN = 'vxlan'
GRE = 'gre'
OVERLAY_NET_TYPES = [VXLAN, GRE]


class ODLControllerRelation(helpers.RelationContext):
    name = 'odl-controller'
    interface = 'odl-controller-api'

    def get_first_data(self):
        if self.get('odl-controller') and len(self['odl-controller']):
            return self['odl-controller'][0]
        else:
            return {}

    def get_data(self):
        super(ODLControllerRelation, self).get_data()
        first_contoller = self.get_first_data()
        # XXX Should be using a VIP (odl doesn't provide one yet) rather than
        # first nodes private-address
        self['odl_ip'] = first_contoller.get('private-address')
        self['odl_port'] = first_contoller.get('port')
        # ODL passing wrong port atm
        self['odl_port'] = '8181'
        self['odl_username'] = first_contoller.get('username')
        self['odl_password'] = first_contoller.get('password')

    def is_ready(self):
        if 'password' in self.get_first_data():
            return True
        else:
            return False


class NeutronPluginRelation(helpers.RelationContext):
    name = 'neutron-plugin'
    interface = 'neutron-plugin-api-subordinate'

    def provide_data(self):
        # Add sections and tuples to insert values into neutron-server's
        # neutron.conf e.g.
        # principle_config = {
        #     "nova-compute": {
        #        "/etc/nova/nova.conf": {
        #             "sections": {
        #                 'DEFAULT': [
        #                     ('key1', 'val1')
        #                     ('key2', 'val2')
        #                 ],
        #                 'libvirt': [
        #                     ('key3', 'val3')
        #                 ],
        #             }
        #         }
        #     }
        # }
        firewall_driver = 'nova.virt.firewall.NoopFirewallDriver'
        vif_driver = 'nova.virt.libvirt.vif.LibvirtGenericVIFDriver'
        principle_config = {
            "nova-compute": {
                "/etc/nova/nova.conf": {
                    "sections": {
                        'DEFAULT': [
                            ('firewall_driver', firewall_driver),
                            ('libvirt_vif_driver', vif_driver),
                            ('security_group_api', 'nova'),
                        ],
                    }
                }
            }
        }
        relation_info = {
            'neutron-plugin': 'odl',
            'subordinate_configuration': json.dumps(principle_config),
        }
        return relation_info


class ConfigTranslation(dict):

    def __init__(self):
        self['use_syslog'] = config('use-syslog')
        self['vlan_ranges'] = config('vlan-ranges')
        self['overlay_network_type'] = self.get_overlay_network_type()

    def get_overlay_network_type(self):
        overlay_networks = config('overlay-network-type').split()
        for overlay_net in overlay_networks:
            if overlay_net not in OVERLAY_NET_TYPES:
                raise ValueError('Unsupported overlay-network-type %s'
                                 % overlay_net)
        return ','.join(overlay_networks)


class PCIInfo(dict):

    def __init__(self):
        ''' Generate pci info '''
        if not self.is_ready():
            log('PCIInfo not ready')
            return
        self.user_requested_config = self.get_user_requested_config()
        net_devices = PCINetDevices()
        self['local_macs'] = net_devices.get_macs()
        pci_addresses = []
        self['local_config'] = {}
        for mac in self.user_requested_config.keys():
            log('Checking if {} is on this host'.format(mac))
            if mac in self['local_macs']:
                log('{} is on this host'.format(mac))
                device = net_devices.get_device_from_mac(mac)
                log('{} is {} and is currently {}'.format(mac,
                    device.pci_address, device.interface_name))
                if device.state == 'up':
                    log('Refusing to add {} to vpe list as it is {}'.format(
                        device.pci_address, device.state))
                else:
                    pci_addresses.append(device.pci_address)
                    self['local_config'][mac] = []
                    for conf in self.user_requested_config[mac]:
                        self['local_config'][mac].append({
                            'net': conf.get('net'),
                            'interface': device.interface_name,
                        })
        if pci_addresses:
            self['pci_devs'] = 'dev ' + ' dev '.join(pci_addresses)
        else:
            self['pci_devs'] = 'no-pci'
        log('pci_devs {}'.format(self['pci_devs']))

    def is_ready(self):
        return os.path.isfile('/opt/cisco/vpe/bin/confd_cli')

    def get_user_requested_config(self):
        ''' Parse the user requested config str mac=<mac>;vlan=<vlan> and
        return a dict'''
        mac_net_config = {}
        mac_map = config('mac-network-map')
        if mac_map:
            for line in mac_map.split():
                entries = line.split(';')
                tmp_dict = {}
                for entry in entries:
                    if '=' in entry:
                        key, value = entry.split('=')
                        tmp_dict[key] = value
                keys = tmp_dict.keys()
                if 'mac' in keys:
                    if tmp_dict['mac'] in mac_net_config:
                        mac_net_config[tmp_dict['mac']].append({
                            'net': tmp_dict.get('net'),
                        })
                    else:
                        mac_net_config[tmp_dict['mac']] = [{
                            'net': tmp_dict.get('net'),
                        }]
        return mac_net_config
