import ODL
import PCIDev
import os
import yaml
import subprocess
import urllib
from charmhelpers.core.sysctl import create as create_sysctl
from socket import gethostname
from charmhelpers.fetch.archiveurl import ArchiveUrlFetchHandler
from charmhelpers.core.hookenv import(
    INFO,
    charm_dir,
    config,
    log,
    unit_private_ip,
)
from charmhelpers.core.hugepage import hugepage_support
from charmhelpers.core.host import (
    fstab_mount,
    rsync,
    user_exists,
)
from charmhelpers.fetch import (
    apt_install,
    apt_purge,
)
import vpp_data
from charmhelpers.core.decorators import retry_on_exception
from charmhelpers.core.services.base import service_restart

ODL_NODE_QUERY_PATH = '/restconf/operational/opendaylight-inventory:nodes/'
ODL_MAC_QUERY_PATH = '/restconf/config/neutron-vpp-map:neutron_net_map/'
ODL_MOUNT_PATH = ('/restconf/config/opendaylight-inventory:nodes/node/'
                  'controller-config/yang-ext:mount/config:modules')

PACKAGES = ['python-psutil', 'gcc', 'make', 'python-crypto', 'qemu', 'seabios']


def install_packages(servicename):
    """ Install base packages """
    purge_packages()
    pkgs = filter_installed_packages(determine_packages())
    apt_install(pkgs)


def install_vpp_packages(servicename):
    """ Install VPP and Extra Kernel modules packages """
    kver = subprocess.check_output(['uname', '-r']).decode('UTF-8').strip()
    extra_pkg = 'linux-image-extra-' + kver
    apt_install(filter_installed_packages([extra_pkg]))
    if filter_installed_packages(['vpe']):
        deb_url = config('vpe-deb-url')
        vpe_deb = charm_dir() + '/payload/vpe.deb'
        if deb_url:
            if deb_url.endswith('.deb'):
                urllib.urlretrieve(deb_url, vpe_deb)
            else:
                handler = ArchiveUrlFetchHandler()
                vpe_deb = handler.install(deb_url) + '/vpe.deb'
        if os.path.isfile(vpe_deb):
            subprocess.check_call(['dpkg', '-i', vpe_deb])


def determine_packages():
    """ Determine which packages are needed for this deployment """
    pkgs = []
    pkgs.extend(PACKAGES)
    if config('use-corekeeper'):
        pkgs.append('corekeeper')
    else:
        pkgs.append('apport')
    return pkgs


def purge_packages():
    """ Remove conflicting packages so required packages can be installed """
    purge_list = []
    crash_manager = ['corekeeper', 'apport']
    white_list = determine_packages()
    black_list = [pkg for pkg in crash_manager if pkg not in white_list]
    not_installed = filter_installed_packages(black_list)
    purge_list.extend([pkg for pkg in black_list if pkg not in not_installed])
    if 'corekeeper' in purge_list:
        bug_1437891()
    apt_purge(purge_list)


def debug_display_qn_pci_conf():
    """ Log VPP config line which includes pci addresses """
    log('pci config from qn.conf')
    with open('/opt/cisco/vpe/etc/qn.conf') as f:
        data = f.readlines()
    for line in data:
        if 'vpe.log' in line:
            log(line)


def restart_qn():
    """ Restart qn (which restarts vpp), log debug and check vpp started """
    log('Restarting qn')
    debug_display_qn_pci_conf()
    service_restart('qn')
    check_vpe_cli()


@retry_on_exception(5, base_delay=10, exc_type=subprocess.CalledProcessError)
def check_vpe_cli():
    """ Query vpp to check it is responding """
    echo_cmd = [
        'echo', '-e', 'show interfaces-state interface phys-address\nexit']
    cli_cmd = ['/opt/cisco/vpe/bin/confd_cli', '-N', '-C', '-u', 'system']
    echo = subprocess.Popen(echo_cmd, stdout=subprocess.PIPE)
    cli_output = subprocess.check_output(cli_cmd, stdin=echo.stdout)
    log('confd_cli returned: {}'.format(cli_output))
    echo.wait()
    echo.terminate


def bug_1437891():
    """ Workaround LP bug 1437891 (dpkg failure when corekeeper is removed) """
    postrm_file = "/var/lib/dpkg/info/corekeeper.postrm"
    lines = []
    with open(postrm_file, 'r') as infile:
        data = infile.readlines()
    for line in data:
        line = line.replace('procps reload', 'procps restart')
        lines.append(line)
    with open(postrm_file, 'w') as outfile:
        for line in lines:
            outfile.write(line)


def configure_ssh_keys(service_name):
    """ Create ssh keys used for talking to ODL"""
    from Crypto.PublicKey import RSA
    # XXX A dsa pub/priv key ships with the vpe package which isn't ideal.
    #     Assume that we can use rsa and ignore the dsa key. Lets get the dsa
    #     key removed from the vpe package anyway.
    key = RSA.generate(2048)
    ssh_dir = '/opt/cisco/vpe/etc/confd/ssh'
    priv_key = ssh_dir + '/ssh_host_rsa_key'
    pub_key = ssh_dir + '/ssh_host_rsa_key.pub'
    if os.path.isfile(priv_key):
        return
    with open(priv_key, 'w') as content_file:
        os.chmod(priv_key, 0600)
        content_file.write(key.exportKey('PEM'))
    pubkey = key.publickey()
    with open(pub_key, 'w') as content_file:
        content_file.write(pubkey.exportKey('OpenSSH'))


def get_odl_config():
    """ Return an ODL connection """
    relation = vpp_data.ODLControllerRelation()
    odl = None
    if relation.is_ready():
        creds = {
            'ip': relation['odl_ip'],
            'port': '8181',
            'username': relation['odl_username'],
            'password': relation['odl_password']
        }
        odl = ODL.ODLConfig(**creds)
        log('Got odl config object')
    else:
        log('Failed to get odl config object')
    return odl


def odl_node_registration(service_name):
    """ Register node with ODL if not registered already """
    odl = get_odl_config()
    device_name = gethostname()
    if odl:
        if odl.is_device_registered(device_name):
            log('{} is already registered in odl'.format(device_name))
        else:
            log('Registering {} ({}) in odl'.format(
                device_name, unit_private_ip()))
            odl.odl_register_node(device_name, unit_private_ip())


def odl_register_macs(service_name):
    """ Register local interfaces and their networks with ODL """
    log('Looking for macs to register with nets in odl')
    odl = get_odl_config()
    device_name = gethostname()
    if odl:
        requested_config = vpp_data.PCIInfo()['local_config']
        for mac in requested_config.keys():
            for requested_net in requested_config[mac]:
                net = requested_net['net']
                interface = requested_net['interface']
                if not odl.is_net_device_registered(net, device_name,
                                                    interface, mac):
                    odl.odl_register_macs(device_name, net, interface, mac)
                    log('Registering {} and {} on '
                        '{}'.format(net, interface, mac))
                else:
                    log('{} already registered for {} on '
                        '{}'.format(net, interface, device_name))


def odl_delete_stale_entries(service_name):
    """ Delete local interface network entries from ODL if no longer needed """
    log('Looking for stale net entries to delete from odl')
    odl = get_odl_config()
    device_name = gethostname()
    if odl:
        pci_info = vpp_data.PCIInfo()
        for mac in pci_info['local_macs']:
            user_requested_conf = pci_info['local_config'].get(mac, [])
            user_requested_nets = [conf['net'] for conf in user_requested_conf]
            log('User has requested {} be bound to '
                '{}'.format(mac, ', '.join(user_requested_nets)))
            for net in odl.get_macs_networks(mac):
                if net not in user_requested_nets:
                    log('Deleting odl net entry for {} on '
                        '{}'.format(device_name, net))
                    odl.delete_net_device_entry(net, device_name)
                else:
                    log('odl net entry for {} on {} is '
                        'legitimate'.format(device_name, net))


def filter_installed_packages(packages, use_dpkg=True):
    """Returns a list of packages that require installation"""
    _pkgs = []
    out = subprocess.check_output(['dpkg', '-l'])
    installed_pkgs = []
    for line in out.split('\n'):
        columns = line.split()
        if len(columns) > 3:
            installed_pkgs.append(columns[1])
    _pkgs = [_pkg for _pkg in packages if _pkg not in installed_pkgs]
    return _pkgs


def vpe_installed():
    """ Return boolean of vpe installed """
    return 'vpe' not in filter_installed_packages('vpe')


def bind_orphaned_net_interfaces(service_name):
    """ If an Ethernet device has no loaded kernel module then bind it back to
    the default for that device type.
    """
    log('Looking for orphans to rebind')
    net_interfaces = PCIDev.PCINetDevices()
    net_interfaces.rebind_orphans()
