#!/usr/bin/python

import amulet
import requests
from neutronclient.v2_0 import client as neutronclient
import apt_pkg

from charmhelpers.contrib.openstack.amulet.deployment import (
    OpenStackAmuletDeployment
)

from charmhelpers.contrib.openstack.amulet.utils import (
    OpenStackAmuletUtils,
    DEBUG,
    ERROR
)

# Use DEBUG to turn on debug logging
u = OpenStackAmuletUtils(DEBUG)

ODL_QUERY_PATH = '/restconf/operational/opendaylight-inventory:nodes/'


class CiscoVPPBasicDeployment(OpenStackAmuletDeployment):
    """Amulet tests on a basic neutron-openvswtich deployment."""

    def __init__(self, series, openstack=None, source=None, git=False,
                 stable=False):
        """Deploy the entire test environment."""
        super(CiscoVPPBasicDeployment, self).__init__(series, openstack,
                                                      source, stable)
        self._add_services()
        self._add_relations()
        self._configure_services()
        self._deploy()
        self._initialize_tests()

    def _add_services(self):
        """Add services

           Add the services that we're testing, where cisco-vpp is local,
           and the rest of the service are from lp branches that are
           compatible with the local charm (e.g. stable or next).
           """
        this_service = {'name': 'cisco-vpp'}
        other_services = [
            {
                'name': 'nova-compute',
                'location':
                'lp:~openstack-charmers/charms/trusty/nova-compute/vpp',
                'constraints': {'mem': '8G'},
            },
            {
                'name': 'neutron-api',
                'location': 'lp:~openstack-charmers/charms/trusty/neutron-api/'
                            'vpp'
            },
            {
                'name': 'odl-controller',
                'location': 'lp:~sdn-charmers/charms/trusty/odl-controller/'
                            'odl-cmds',
                'constraints': {'mem': '8G'},
            },
            {
                'name': 'neutron-api-odl',
                'location': 'lp:~openstack-charmers/charms/trusty/'
                            'neutron-api-odl/vpp',
            },
            {'name': 'mysql'},
            {'name': 'rabbitmq-server'},
            {'name': 'keystone'},
            {'name': 'nova-cloud-controller'},
            {'name': 'glance'},
        ]
        super(CiscoVPPBasicDeployment, self)._add_services(this_service,
                                                           other_services)

    def _add_relations(self):
        """Add all of the relations for the services."""
        relations = {
            'neutron-api:neutron-plugin-api-subordinate':
            'neutron-api-odl:neutron-plugin-api-subordinate',
            'neutron-api:neutron-plugin-api':
            'cisco-vpp:neutron-plugin-api',
            'nova-compute:neutron-plugin': 'cisco-vpp:neutron-plugin',
            'cisco-vpp:odl-controller': 'odl-controller:controller-api',
            'neutron-api-odl:odl-controller': 'odl-controller:controller-api',
            'keystone:shared-db': 'mysql:shared-db',
            'nova-cloud-controller:shared-db': 'mysql:shared-db',
            'nova-cloud-controller:amqp': 'rabbitmq-server:amqp',
            'nova-cloud-controller:image-service': 'glance:image-service',
            'nova-cloud-controller:identity-service':
            'keystone:identity-service',
            'nova-compute:cloud-compute':
            'nova-cloud-controller:cloud-compute',
            'nova-compute:amqp': 'rabbitmq-server:amqp',
            'nova-compute:image-service': 'glance:image-service',
            'glance:shared-db': 'mysql:shared-db',
            'glance:identity-service': 'keystone:identity-service',
            'glance:amqp': 'rabbitmq-server:amqp',
            'neutron-api:shared-db': 'mysql:shared-db',
            'neutron-api:amqp': 'rabbitmq-server:amqp',
            'neutron-api:neutron-api': 'nova-cloud-controller:neutron-api',
            'neutron-api:identity-service': 'keystone:identity-service',
        }
        super(CiscoVPPBasicDeployment, self)._add_relations(relations)

    def _configure_services(self):
        """Configure all of the services."""
        resource_addr = 'http://10.5.17.29/'
        odl_binary = 'distribution-karaf-1.0.0-00004.1-SNAPSHOT.tar.gz'
        neutron_api = {
            'manage-neutron-plugin-legacy-mode': False,
            'neutron-security-groups': False,
            'overlay-network-type': 'vxlan',
        }
        nova_compute = {
            'enable-live-migration': False,
            'manage-neutron-plugin-legacy-mode': False,
            'openstack-origin': 'ppa:sdn-charmers/cisco-vpp-testing',
        }
        keystone = {
            'admin-password': 'openstack',
            'admin-token': 'ubuntutesting',
        }
        mysql = {
            'dataset-size': '50%',
        }
        odl_controller = {
            'install-url': resource_addr + odl_binary,
        }
        cisco_vpp = {
            'vpe-deb-url': resource_addr + '/vpe.deb.tgz',
            'overlay-network-type': 'vxlan',
        }
        neutron_api_odl = {
            'overlay-network-type': 'vxlan',
        }
        configs = {
            'neutron-api': neutron_api,
            'nova-compute': nova_compute,
            'keystone': keystone,
            'mysql': mysql,
            'odl-controller': odl_controller,
            'cisco-vpp': cisco_vpp,
            'neutron-api-odl': neutron_api_odl,
        }
        super(CiscoVPPBasicDeployment, self)._configure_services(configs)

    def _initialize_tests(self):
        """Perform final initialization before tests get run."""
        # Access the sentries for inspecting service units
        self.compute_sentry = self.d.sentry.unit['nova-compute/0']
        self.neutron_api_sentry = self.d.sentry.unit['neutron-api/0']
        self.ciscovpp_sentry = self.d.sentry.unit['cisco-vpp/0']
        self.mysql_sentry = self.d.sentry.unit['mysql/0']
        self.rabbitmq_server_sentry = self.d.sentry.unit['rabbitmq-server/0']
        self.keystone_sentry = self.d.sentry.unit['keystone/0']
        self.glance_sentry = self.d.sentry.unit['glance/0']
        self.nova_cc_sentry = self.d.sentry.unit['nova-cloud-controller/0']
        self.neutron_api_odl_sentry = self.d.sentry.unit['neutron-api-odl/0']
        self.odl_controller_sentry = self.d.sentry.unit['odl-controller/0']
        self.keystone = u.authenticate_keystone_admin(self.keystone_sentry,
                                                      user='admin',
                                                      password='openstack',
                                                      tenant='admin')
        ep = self.keystone.service_catalog.url_for(service_type='identity',
                                                   endpoint_type='publicURL')
        self.neutron = neutronclient.Client(auth_url=ep,
                                            username='admin',
                                            password='openstack',
                                            tenant_name='admin',
                                            region_name='RegionOne')

    def test_services(self):
        """Verify the expected services are running on the corresponding
           service units."""

        commands = {
            self.compute_sentry: ['status nova-compute'],
        }

        ret = u.validate_services(commands)
        if ret:
            amulet.raise_status(amulet.FAIL, msg=ret)

    def test_check_registration(self):
        unit = self.odl_controller_sentry
        comp_hostname = self.compute_sentry.run('hostname -s')[0]
        odl_ip = unit.relation(
            'controller-api',
            'neutron-api-odl:odl-controller'
        )['private-address']

        odl_base_url = 'http://{}:{}'.format(odl_ip, '8181')
        query_url = odl_base_url + ODL_QUERY_PATH
        odl_creds = ('admin', 'admin')
        odl_req = requests.get(
            query_url,
            auth=odl_creds,
            proxies={},
            timeout=10
        )
        if odl_req.status_code == 200:
            nodes = [entry['id'] for entry in odl_req.json()[u'nodes']['node']]
            if comp_hostname not in nodes:
                msg = 'Compute node not registered with odl controller'
                amulet.raise_status(amulet.FAIL, msg=msg)
        else:
            amulet.raise_status(amulet.FAIL, msg='ODL query failed')

    def test_net_create(self):
        net_name = 'net_d1_164'
        old_net = self.neutron.list_networks(name=net_name)
        if old_net['networks']:
            old_net_id = old_net['networks'][0]['id']
            self.neutron.delete_network(old_net_id)
        network = {
            'name': net_name,
            'provider:network_type': 'vlan',
            'provider:physical_network': 'net_d1',
            'provider:segmentation_id': 164,
        }
        self.neutron.create_network({'network': network})
        new_nets = self.neutron.list_networks(name=net_name)
        if not new_nets['networks']:
            amulet.raise_status(amulet.FAIL, msg='New net not found')

    def test_check_noovs_services(self):
        check_cmd = ('{ sudo service neutron-plugin-openvswitch-agent status'
                     '|| sudo service neutron-ovs-cleanup status; } || exit 0')
        output, rc = self.compute_sentry.run(check_cmd)
        if rc != 0:
            msg = 'OVS service found (should be absent)'
            amulet.raise_status(amulet.FAIL, msg=msg)

    def test_check_no_bridges(self):
        check_cmd = ("ip addr list | awk '/br-/ {print;}'")
        output, rc = self.compute_sentry.run(check_cmd)
        if output:
            msg = 'Bridges found (should be absent)'
            amulet.raise_status(amulet.FAIL, msg=msg)

    def test_check_no_ovs_pkgs(self):
        check_cmd = ("{ dpkg -l | grep -E 'openvswitch-switch|"
                     "neutron-plugin-openvswitch-agent'; } || exit 0;")
        output, rc = self.compute_sentry.run(check_cmd)
        if output:
            msg = 'ovs packages found (should be absent)'
            amulet.raise_status(amulet.FAIL, msg=msg)

    def test_check_no_ovs_kmods(self):
        check_cmd = ("{ lsmod | grep 'openswitch'; } || exit 0;")
        output, rc = self.compute_sentry.run(check_cmd)
        if output:
            msg = 'ovs kernel module found (should be absent)'
            amulet.raise_status(amulet.FAIL, msg=msg)

    def test_hugepage_kernel_params(self):
        check_cmd = ("sysctl -a | grep -E '(nr_hugepages|max_map_count|"
                     "hugetlb_shm_group|shmmax)\s'")
        output, rc = self.compute_sentry.run(check_cmd)
        expect = {
            'vm.max_map_count': 3072,
            'vm.nr_hugepages': 1024,
            'vm.hugetlb_shm_group': 0,
            'kernel.shmmax': 2147483648,
        }
        for line in output.split('\n'):
            columns = line.split()
            if expect[columns[0]] < int(columns[2]):
                msg = 'Kernel setting {} too low ({} < {})'.format(
                    columns[0],
                    expect[columns[0]],
                    columns[2]
                )
                amulet.raise_status(amulet.FAIL, msg=msg)

    def test_hugepage_sysctld_settings(self):
        unit = self.compute_sentry
        shmmax_conf = unit.file_contents('/etc/sysctl.d/10-shmmax.conf')
        hugepage_conf = unit.file_contents('/etc/sysctl.d/10-hugepage.conf')
        expect = {
            'vm.max_map_count': 3072,
            'vm.nr_hugepages': 1024,
            'vm.hugetlb_shm_group': 0,
            'kernel.shmmax': 2147483648,
        }
        for line in shmmax_conf.split('\n') + hugepage_conf.split('\n'):
            columns = line.split('=')
            if not columns[0]:
                continue
            if expect[columns[0]] < int(columns[1]):
                msg = 'Kernel setting {} too low ({} < {})'.format(
                    columns[0],
                    expect[columns[0]],
                    columns[1],
                )
                amulet.raise_status(amulet.FAIL, msg=msg)

    def test_hugepage_mount_point(self):
        unit = self.compute_sentry
        mtab = unit.file_contents('/etc/mtab')
        mnt_entry = [line for line in mtab.split('\n') if '/mnt/huge' in line]
        if not mnt_entry:
            msg = 'Hugepage fs not found in mtab'
            amulet.raise_status(amulet.FAIL, msg=msg)

    def test_hugepage_init_script(self):
        unit = self.compute_sentry
        unit.file('/etc/init.d/qemu-hugefsdir')

    def test_vpe_running(self):
        check_cmd = ("pgrep vpe")
        output, rc = self.compute_sentry.run(check_cmd)
        if rc != 0:
            msg = 'vpe does not appear to be running'
            amulet.raise_status(amulet.FAIL, msg=msg)

    def test_ssh_files(self):
        unit = self.compute_sentry
        unit.file('/cisco/etc/confd/ssh/ssh_host_dsa_key')
        unit.file('/cisco/etc/confd/ssh/ssh_host_dsa_key.pub')
        unit.file('/cisco/etc/confd/ssh/ssh_host_rsa_key')
        unit.file('/cisco/etc/confd/ssh/ssh_host_rsa_key.pub')

    def test_pkg_version(self):
        sentry = self.compute_sentry
        apt_pkg.init()
        dpkg_query = ("dpkg-query --show --showformat '${Version}'")
        qemu_revno, qrc = sentry.run(dpkg_query + ' qemu')
        libvirt_revno, qrc = sentry.run(dpkg_query + ' libvirt-bin')
        qemu_cmp_rc = apt_pkg.version_compare(qemu_revno, '1:2.2')
        libvirt_cmp_rc = apt_pkg.version_compare(libvirt_revno, '1.2.8')
        if libvirt_cmp_rc < 0 or qemu_cmp_rc < 0:
            amulet.raise_status(amulet.FAIL, msg='pkg version too low')

    def parse_config(self, conf_contents):
        settings = {}
        for line in conf_contents.split('\n'):
            if '=' in line and not line.startswith('#'):
                line = line.replace(' ', '')
                key, value = line.split('=')
                settings[key] = value
        return settings

    def test_ml2_config(self):
        unit = self.compute_sentry
        ml2_conf = unit.file_contents('/etc/neutron/plugins/ml2/ml2_conf.ini')
        expected = {
            'type_drivers': 'vxlan',
            'username': 'admin',
            'mechanism_drivers': 'opendaylight',
            'url': 'http://10.5.34.203:8080/controller/nb/v2/neutron',
            'firewall_driver':
            'neutron.agent.linux.firewall.NoopFirewallDriver',
            'enable_security_group': 'False',
            'vni_ranges': '1001:2000',
            'tenant_network_types': 'gre,vxlan,vlan,flat',
            'password': 'admin',
            'tunnel_id_ranges': '1:1000',
        }
        settings = self.parse_config(ml2_conf)
        for key in expected.keys():
            if expected[key] != settings[key]:
                msg = 'Unexpected value in ml2_conf.ini'
                amulet.raise_status(amulet.FAIL, msg=msg)

    def test_libvirt_config(self):
        unit = self.compute_sentry
        libvirt_conf = unit.file_contents('/etc/libvirt/libvirtd.conf')
        expected = {
            'listen_tls': '0',
        }
        settings = self.parse_config(libvirt_conf)
        for key in expected.keys():
            if expected[key] != settings[key]:
                msg = 'Unexpected value in libvirtd.conf'
                amulet.raise_status(amulet.FAIL, msg=msg)

    def test_nova_config(self):
        unit = self.compute_sentry
        nova_conf = unit.file_contents('/etc/nova/nova.conf')
        expected = {
            'security_group_api': 'nova',
            'firewall_driver': 'nova.virt.firewall.NoopFirewallDriver',
        }
        settings = self.parse_config(nova_conf)
        for key in expected.keys():
            if expected[key] != settings[key]:
                msg = 'Unexpected value in nova.conf'
                amulet.raise_status(amulet.FAIL, msg=msg)
