from test_utils import CharmTestCase, patch_open
from test_pcidev_helper import (
    check_device,
    mocked_subprocess,
    mocked_filehandle,
    mocked_globs,
    mocked_islink,
    mocked_realpath,
)
from mock import patch, MagicMock
import PCIDev
import nose.tools
import pci_responses

TO_PATCH = [
    'glob',
    'log',
    'subprocess',
]
NOT_JSON = "Im not json"


class PCIDevTest(CharmTestCase):

    def setUp(self):
        super(PCIDevTest, self).setUp(PCIDev, TO_PATCH)

    def tearDown(self):
        super(PCIDevTest, self).tearDown()

    def test_format_pci_addr(self):
        self.assertEqual(PCIDev.format_pci_addr('0:0:1.1'), '0000:00:01.1')
        self.assertEqual(PCIDev.format_pci_addr(
            '0000:00:02.1'), '0000:00:02.1')


class PCINetDeviceTest(CharmTestCase):

    def setUp(self):
        super(PCINetDeviceTest, self).setUp(PCIDev, TO_PATCH)

    def tearDown(self):
        super(PCINetDeviceTest, self).tearDown()

    @patch('os.path.islink')
    @patch('os.path.realpath')
    def eth_int(self, pci_address, _osrealpath, _osislink, subproc_map=None):
        self.glob.glob.side_effect = mocked_globs
        _osislink.side_effect = mocked_islink
        _osrealpath.side_effect = mocked_realpath
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=subproc_map)

        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            _file.readlines.side_effect = super_fh._getfilecontents_readlines
            netint = PCIDev.PCINetDevice(pci_address)
        return netint

    def test_base_eth_device(self):
        net = self.eth_int('0000:10:00.0')
        expect = {
            'interface_name': 'eth2',
            'mac_address': 'a8:9d:21:cf:93:fc',
            'pci_address': '0000:10:00.0',
            'modalias_kmod': 'igb',
            'loaded_kmod': 'igb',
            'state': 'up',
        }
        self.assertTrue(check_device(net, expect))

    def test_base_vpe_device(self):
        net = self.eth_int('0000:06:00.0')
        expect = {
            'interface_name': 'TenGigabitEthernet6/0/0',
            'mac_address': '84:b8:02:2a:5f:c3',
            'pci_address': '0000:06:00.0',
            'modalias_kmod': 'enic',
            'loaded_kmod': 'igb_uio',
            'state': 'vpebound',
        }
        self.assertTrue(check_device(net, expect))

    def test_base_orphan(self):
        net = self.eth_int(
            '0000:07:00.0',
            subproc_map=pci_responses.NET_SETUP_ORPHAN
        )
        expect = {
            'interface_name': None,
            'mac_address': None,
            'pci_address': '0000:07:00.0',
            'modalias_kmod': 'enic',
            'loaded_kmod': None,
            'state': 'unbound',
        }
        self.assertTrue(check_device(net, expect))

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_loaded_kmod(self, _update):
        device = PCIDev.PCINetDevice('0000:06:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess()
        device.update_loaded_kmod()
        self.assertEqual(device.loaded_kmod, 'igb_uio')

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_loaded_kmod_notloaded(self, _update):
        device = PCIDev.PCINetDevice('0000:07:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=pci_responses.NET_SETUP_ORPHAN)
        device.update_loaded_kmod()
        self.assertEqual(device.loaded_kmod, None)

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_modalias_kmod(self, _update):
        device = PCIDev.PCINetDevice('0000:07:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess()
        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            _file.readlines.side_effect = super_fh._getfilecontents_readlines
            device.update_modalias_kmod()
        self.assertEqual(device.modalias_kmod, 'enic')

    @patch('PCIDev.PCINetDevice.update_interface_info_eth')
    @patch('PCIDev.PCINetDevice.update_interface_info_vpe')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_interface_info_call_vpeinfo(self, _update, _vpe_info,
                                                _eth_info):
        dev6 = PCIDev.PCINetDevice('0000:06:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess()
        dev6.update_loaded_kmod()
        dev6.update_interface_info()
        _vpe_info.assert_called_with()
        self.assertFalse(_eth_info.called)

    @patch('PCIDev.PCINetDevice.update_interface_info_eth')
    @patch('PCIDev.PCINetDevice.update_interface_info_vpe')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_interface_info_call_ethinfo(self, _update, _vpe_info,
                                                _eth_info):
        dev = PCIDev.PCINetDevice('0000:10:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess()
        dev.update_loaded_kmod()
        dev.update_interface_info()
        _eth_info.assert_called_with()
        self.assertFalse(_vpe_info.called)

    @patch('PCIDev.PCINetDevice.update_interface_info_eth')
    @patch('PCIDev.PCINetDevice.update_interface_info_vpe')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_interface_info_orphan(self, _update, _vpe_info, _eth_info):
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=pci_responses.NET_SETUP_ORPHAN)
        dev.update_loaded_kmod()
        dev.update_interface_info()
        self.assertFalse(_vpe_info.called)
        self.assertFalse(_eth_info.called)
        self.assertEqual(dev.interface_name, None)
        self.assertEqual(dev.mac_address, None)
        self.assertEqual(dev.state, 'unbound')

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_kernel_name(self, _update):
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        self.subprocess.check_output.return_value = '3.13.0-55-generic'
        self.assertEqual(dev.get_kernel_name(), '3.13.0-55-generic')

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_pci_rescan(self, _update):
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        with patch_open() as (_open, _file):
            dev.pci_rescan()
            _open.assert_called_with('/sys/bus/pci/rescan', 'w')
            _file.write.assert_called_with('1')

    @patch('PCIDev.PCINetDevice.pci_rescan')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_bind(self, _update, _pci_rescan):
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        self.subprocess.check_output.side_effect = mocked_subprocess()
        with patch_open() as (_open, _file):
            dev.bind('enic')
            _open.assert_called_with('/sys/bus/pci/drivers/enic/bind', 'w')
            _file.write.assert_called_with('0000:07:00.0')
        _pci_rescan.assert_called_with()
        _update.assert_called_with()

    @patch('PCIDev.PCINetDevice.pci_rescan')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_unbind(self, _update, _pci_rescan):
        self.subprocess.check_output.side_effect = mocked_subprocess()
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        dev.update_loaded_kmod()
        with patch_open() as (_open, _file):
            dev.unbind()
            _open.assert_called_with(
                '/sys/bus/pci/drivers/igb_uio/unbind', 'w')
            _file.write.assert_called_with('0000:07:00.0')
        _pci_rescan.assert_called_with()
        _update.assert_called_with()

    @patch('PCIDev.PCINetDevice.pci_rescan')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_unbind_nokmod(self, _update, _pci_rescan):
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=pci_responses.NET_SETUP_ORPHAN)
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        dev.update_loaded_kmod()
        with patch_open() as (_open, _file):
            dev.unbind()
            self.assertFalse(_open.called)
        self.assertFalse(_pci_rescan.called)

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_interface_info_vpe(self, _update):
        self.subprocess.check_output.side_effect = mocked_subprocess()
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        dev.update_interface_info_vpe()
        self.assertEqual('TenGigabitEthernet7/0/0', dev.interface_name)
        self.assertEqual('84:b8:02:2a:5f:c4', dev.mac_address)
        self.assertEqual('vpebound', dev.state)

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_interface_info_vpe_orphan(self, _update):
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=pci_responses.NET_SETUP_ORPHAN)
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        dev.update_interface_info_vpe()
        self.assertEqual(None, dev.interface_name)
        self.assertEqual(None, dev.mac_address)
        self.assertEqual(None, dev.state)

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_vpe_cli_out(self, _update):
        self.subprocess.check_output.side_effect = mocked_subprocess()
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        self.assertTrue('local0' in dev.get_vpe_cli_out())

    @patch('PCIDev.PCINetDevice.get_vpe_cli_out')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_vpe_interfaces_and_macs(self, _update, _get_vpe_cli_out):
        self.subprocess.check_output.side_effect = mocked_subprocess()
        _get_vpe_cli_out.return_value = pci_responses.CONFD_CLI
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        vpe_devs = dev.get_vpe_interfaces_and_macs()
        expect = [
            {
                'interface': 'TenGigabitEthernet6/0/0',
                'macAddress': '84:b8:02:2a:5f:c3',
                'pci_address': '0000:06:00.0'
            },
            {
                'interface': 'TenGigabitEthernet7/0/0',
                'macAddress': '84:b8:02:2a:5f:c4',
                'pci_address': '0000:07:00.0'
            },
        ]
        self.assertEqual(vpe_devs, expect)

    @patch('PCIDev.PCINetDevice.get_vpe_cli_out')
    @patch('PCIDev.PCINetDevice.update_attributes')
    @patch('subprocess.CalledProcessError')
    def test_get_vpe_interfaces_and_macs_invalid_cli(self, _excpt, _update,
                                                     _get_vpe_cli_out):
        self.subprocess.check_output.side_effect = mocked_subprocess()
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        _get_vpe_cli_out.return_value = pci_responses.CONFD_CLI_NOLOCAL
        with nose.tools.assert_raises(Exception):
            dev.get_vpe_interfaces_and_macs()

    @patch('PCIDev.PCINetDevice.get_vpe_cli_out')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_vpe_interfaces_and_macs_invmac(self, _update,
                                                _get_vpe_cli_out):
        self.subprocess.check_output.side_effect = mocked_subprocess()
        _get_vpe_cli_out.return_value = pci_responses.CONFD_CLI_INVMAC
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        vpe_devs = dev.get_vpe_interfaces_and_macs()
        expect = [
            {
                'interface': 'TenGigabitEthernet7/0/0',
                'macAddress': '84:b8:02:2a:5f:c4',
                'pci_address': '0000:07:00.0'
            },
        ]
        self.assertEqual(vpe_devs, expect)

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_extract_pci_addr_from_nic(self, _update):
        dev = PCIDev.PCINetDevice('0000:07:00.0')
        self.assertEqual(dev.extract_pci_addr_from_vpe_interface(
            'TenGigabitEthernet1/1/1'), '0000:01:01.1')
        self.assertEqual(dev.extract_pci_addr_from_vpe_interface(
            'TenGigabitEtherneta/0/0'), '0000:0a:00.0')
        self.assertEqual(dev.extract_pci_addr_from_vpe_interface(
            'GigabitEthernet0/2/0'), '0000:00:02.0')

    @patch('PCIDev.PCINetDevice.get_sysnet_interfaces_and_macs')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_interface_info_eth(self, _update, _sysnet_ints):
        dev = PCIDev.PCINetDevice('0000:10:00.0')
        _sysnet_ints.return_value = [
            {
                'interface': 'eth2',
                'macAddress': 'a8:9d:21:cf:93:fc',
                'pci_address': '0000:10:00.0',
                'state': 'up'
            },
            {
                'interface': 'eth3',
                'macAddress': 'a8:9d:21:cf:93:fd',
                'pci_address': '0000:10:00.1',
                'state': 'down'
            }
        ]
        dev.update_interface_info_eth()
        self.assertEqual(dev.interface_name, 'eth2')

    @patch('os.path.islink')
    @patch('os.path.realpath')
    @patch('PCIDev.PCINetDevice.get_sysnet_device_state')
    @patch('PCIDev.PCINetDevice.get_sysnet_mac')
    @patch('PCIDev.PCINetDevice.get_sysnet_interface')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_sysnet_interfaces_and_macs(self, _update, _interface, _mac,
                                            _state, _osrealpath, _osislink):
        dev = PCIDev.PCINetDevice('0000:06:00.0')
        self.glob.glob.return_value = ['/sys/class/net/eth2']
        _interface.return_value = 'eth2'
        _mac.return_value = 'a8:9d:21:cf:93:fc'
        _state.return_value = 'up'
        _osrealpath.return_value = ('/sys/devices/pci0000:00/0000:00:02.0/'
                                    '0000:02:00.0/0000:03:00.0/0000:04:00.0/'
                                    '0000:05:01.0/0000:07:00.0')
        expect = {
            'interface': 'eth2',
            'macAddress': 'a8:9d:21:cf:93:fc',
            'pci_address': '0000:07:00.0',
            'state': 'up',
        }
        self.assertEqual(dev.get_sysnet_interfaces_and_macs(), [expect])

    @patch('os.path.islink')
    @patch('os.path.realpath')
    @patch('PCIDev.PCINetDevice.get_sysnet_device_state')
    @patch('PCIDev.PCINetDevice.get_sysnet_mac')
    @patch('PCIDev.PCINetDevice.get_sysnet_interface')
    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_sysnet_interfaces_and_macs_virtio(self, _update, _interface,
                                                   _mac, _state, _osrealpath,
                                                   _osislink):
        dev = PCIDev.PCINetDevice('0000:06:00.0')
        self.glob.glob.return_value = ['/sys/class/net/eth2']
        _interface.return_value = 'eth2'
        _mac.return_value = 'a8:9d:21:cf:93:fc'
        _state.return_value = 'up'
        _osrealpath.return_value = ('/sys/devices/pci0000:00/0000:00:07.0/'
                                    'virtio5')
        expect = {
            'interface': 'eth2',
            'macAddress': 'a8:9d:21:cf:93:fc',
            'pci_address': '0000:00:07.0',
            'state': 'up',
        }
        self.assertEqual(dev.get_sysnet_interfaces_and_macs(), [expect])

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_sysnet_mac(self, _update):
        device = PCIDev.PCINetDevice('0000:10:00.1')
        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            macaddr = device.get_sysnet_mac('/sys/class/net/eth3')
        self.assertEqual(macaddr, 'a8:9d:21:cf:93:fd')

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_sysnet_device_state(self, _update):
        device = PCIDev.PCINetDevice('0000:10:00.1')
        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            state = device.get_sysnet_device_state('/sys/class/net/eth3')
        self.assertEqual(state, 'down')

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_get_sysnet_interface(self, _update):
        device = PCIDev.PCINetDevice('0000:10:00.1')
        self.assertEqual(
            device.get_sysnet_interface('/sys/class/net/eth3'), 'eth3')


class PCINetDevicesTest(CharmTestCase):

    def setUp(self):
        super(PCINetDevicesTest, self).setUp(PCIDev, TO_PATCH)

    def tearDown(self):
        super(PCINetDevicesTest, self).tearDown()

    @patch('os.path.islink')
    def pci_devs(self, _osislink, subproc_map=None):
        self.glob.glob.side_effect = mocked_globs
        rp_patcher = patch('os.path.realpath')
        rp_mock = rp_patcher.start()
        rp_mock.side_effect = mocked_realpath
        _osislink.side_effect = mocked_islink
        self.subprocess.check_output.side_effect = mocked_subprocess(
            subproc_map=subproc_map)

        with patch_open() as (_open, _file):
            super_fh = mocked_filehandle()
            _file.readlines = MagicMock()
            _open.side_effect = super_fh._setfilename
            _file.read.side_effect = super_fh._getfilecontents_read
            _file.readlines.side_effect = super_fh._getfilecontents_readlines
            devices = PCIDev.PCINetDevices()
        rp_patcher.stop()
        return devices

    def test_base(self):
        devices = self.pci_devs()
        self.assertEqual(len(devices.pci_devices), 4)
        expect = {
            '0000:06:00.0': {
                'interface_name': 'TenGigabitEthernet6/0/0',
                'mac_address': '84:b8:02:2a:5f:c3',
                'pci_address': '0000:06:00.0',
                'modalias_kmod': 'enic',
                'loaded_kmod': 'igb_uio',
                'state': 'vpebound',
            },
            '0000:07:00.0': {
                'interface_name': 'TenGigabitEthernet7/0/0',
                'mac_address': '84:b8:02:2a:5f:c4',
                'pci_address': '0000:07:00.0',
                'modalias_kmod': 'enic',
                'loaded_kmod': 'igb_uio',
                'state': 'vpebound',
            },
            '0000:10:00.0': {
                'interface_name': 'eth2',
                'mac_address': 'a8:9d:21:cf:93:fc',
                'pci_address': '0000:10:00.0',
                'modalias_kmod': 'igb',
                'loaded_kmod': 'igb',
                'state': 'up',
            },
            '0000:10:00.1': {
                'interface_name': 'eth3',
                'mac_address': 'a8:9d:21:cf:93:fd',
                'pci_address': '0000:10:00.1',
                'modalias_kmod': 'igb',
                'loaded_kmod': 'igb',
                'state': 'down',
            },
        }
        for device in devices.pci_devices:
            self.assertTrue(check_device(device, expect[device.pci_address]))

    def test_base_orphan(self):
        devices = self.pci_devs(subproc_map=pci_responses.NET_SETUP_ORPHAN)
        expect = {
            'interface_name': None,
            'mac_address': None,
            'pci_address': '0000:07:00.0',
            'modalias_kmod': 'enic',
            'loaded_kmod': None,
            'state': 'unbound',
        }
        device = devices.get_device_from_pci_address('0000:07:00.0')
        self.assertTrue(check_device(device, expect))

    def test_get_pci_ethernet_addresses(self):
        devices = self.pci_devs()
        expect = ['0000:06:00.0', '0000:07:00.0', '0000:10:00.0',
                  '0000:10:00.1']
        self.assertEqual(devices.get_pci_ethernet_addresses(), expect)

    @patch('PCIDev.PCINetDevice.update_attributes')
    def test_update_devices(self, _update):
        devices = self.pci_devs()
        call_count = _update.call_count
        devices.update_devices()
        self.assertEqual(_update.call_count, call_count + 4)

    def test_get_macs(self):
        devices = self.pci_devs()
        expect = ['84:b8:02:2a:5f:c3', '84:b8:02:2a:5f:c4',
                  'a8:9d:21:cf:93:fc', 'a8:9d:21:cf:93:fd']
        self.assertEqual(devices.get_macs(), expect)

    def test_get_macs_orphans(self):
        devices = self.pci_devs(subproc_map=pci_responses.NET_SETUP_ORPHAN)
        expect = ['84:b8:02:2a:5f:c3', 'a8:9d:21:cf:93:fc',
                  'a8:9d:21:cf:93:fd']
        self.assertEqual(devices.get_macs(), expect)

    def test_get_device_from_mac(self):
        devices = self.pci_devs()
        expect = {
            '0000:06:00.0': {
                'interface_name': 'TenGigabitEthernet6/0/0',
                'mac_address': '84:b8:02:2a:5f:c3',
                'pci_address': '0000:06:00.0',
                'modalias_kmod': 'enic',
                'loaded_kmod': 'igb_uio',
                'state': 'vpebound',
            },
            '0000:10:00.1': {
                'interface_name': 'eth3',
                'mac_address': 'a8:9d:21:cf:93:fd',
                'pci_address': '0000:10:00.1',
                'modalias_kmod': 'igb',
                'loaded_kmod': 'igb',
                'state': 'down',
            },
        }
        self.assertTrue(check_device(
            devices.get_device_from_mac('84:b8:02:2a:5f:c3'),
            expect['0000:06:00.0']))
        self.assertTrue(check_device(
            devices.get_device_from_mac('a8:9d:21:cf:93:fd'),
            expect['0000:10:00.1']))

    def test_get_device_from_pci_address(self):
        devices = self.pci_devs()
        expect = {
            '0000:06:00.0': {
                'interface_name': 'TenGigabitEthernet6/0/0',
                'mac_address': '84:b8:02:2a:5f:c3',
                'pci_address': '0000:06:00.0',
                'modalias_kmod': 'enic',
                'loaded_kmod': 'igb_uio',
                'state': 'vpebound',
            },
            '0000:10:00.1': {
                'interface_name': 'eth3',
                'mac_address': 'a8:9d:21:cf:93:fd',
                'pci_address': '0000:10:00.1',
                'modalias_kmod': 'igb',
                'loaded_kmod': 'igb',
                'state': 'down',
            },
        }
        self.assertTrue(check_device(
            devices.get_device_from_pci_address('0000:06:00.0'),
            expect['0000:06:00.0']))
        self.assertTrue(check_device(
            devices.get_device_from_pci_address('0000:10:00.1'),
            expect['0000:10:00.1']))

    @patch('PCIDev.PCINetDevices.unbind_orphans')
    @patch('PCIDev.PCINetDevices.bind_orphans')
    def test_rebind_orphans(self, _bind, _unbind):
        devices = self.pci_devs()
        devices.rebind_orphans()
        _bind.assert_called_with()
        _unbind.assert_called_with()

    @patch('PCIDev.PCINetDevices.update_devices')
    @patch('PCIDev.PCINetDevice.unbind')
    def test_unbind_orphans(self, _unbind, _update_devices):
        devices = self.pci_devs(subproc_map=pci_responses.NET_SETUP_ORPHAN)
        devices.unbind_orphans()
        _unbind.assert_called_with()
        _update_devices.assert_called_with()

    @patch('PCIDev.PCINetDevices.update_devices')
    @patch('PCIDev.PCINetDevice.unbind')
    def test_unbind_orphans_noorphans(self, _unbind, _update_devices):
        devices = self.pci_devs()
        devices.unbind_orphans()
        self.assertEqual(_unbind.call_args_list, [])
        _update_devices.assert_called_with()

    @patch('PCIDev.PCINetDevices.update_devices')
    @patch('PCIDev.PCINetDevice.bind')
    def test_bind_orphans(self, _bind, _update_devices):
        devices = self.pci_devs(subproc_map=pci_responses.NET_SETUP_ORPHAN)
        devices.bind_orphans()
        _bind.assert_called_with('enic')
        _update_devices.assert_called_with()

    @patch('PCIDev.PCINetDevices.update_devices')
    @patch('PCIDev.PCINetDevice.bind')
    def test_bind_orphans_noorphans(self, _bind, _update_devices):
        devices = self.pci_devs()
        devices.bind_orphans()
        self.assertEqual(_bind.call_args_list, [])
        _update_devices.assert_called_with()

    def test_get_orphans(self):
        devices = self.pci_devs(subproc_map=pci_responses.NET_SETUP_ORPHAN)
        expect = {
            'interface_name': None,
            'mac_address': None,
            'pci_address': '0000:07:00.0',
            'modalias_kmod': 'enic',
            'loaded_kmod': None,
            'state': 'unbound',
        }
        orphans = devices.get_orphans()
        self.assertEqual(len(orphans), 1)
        self.assertTrue(check_device(orphans[0], expect))
