from test_utils import CharmTestCase
from mock import patch
import vpp_data
import charmhelpers
import charmhelpers.core.services.helpers
import nose.tools
import json
import test_pcidev_helper

TO_PATCH = [
    'PCINetDevices',
    'config',
    'log',
]


def fake_context(settings):
    def outer():
        def inner():
            return settings
        return inner
    return outer


FULL_NPAPI = {
    'data': {
        'overlay-network-type': 'gre',
        'network-device-mtu': '1200',
    },
    'rids': ['neutron-plugin-api:2'],
    'runits': ['neutron-api/0'],
}

NOMTU_NPAPI = {
    'data': {
        'overlay-network-type': 'gre',
    },
    'rids': ['neutron-plugin-api:2'],
    'runits': ['neutron-api/0'],
}

MISSING_DATA_NPAPI = {
    'data': {},
    'rids': ['neutron-plugin-api:2'],
    'runits': ['neutron-api/0'],
}

FULL_ODLCTRL = {
    'data': {
        'private-address': '10.0.0.27',
        'port': '8080',
        'username': 'odluser',
        'password': 'hardpassword',
    },
    'rids': ['odl-controller:2'],
    'runits': ['odl-controller/0'],
}

MISSING_DATA_ODLCTRL = {
    'data': {
        'private-address': '10.0.0.27',
    },
    'rids': ['odl-controller:2'],
    'runits': ['odl-controller/0'],
}
EMPTY_ODLCTRL = {
    'data': {},
    'rids': [],
    'runits': [],
}

NP_REL = {
    'data': {},
    'rids': ['neutron-plugin:9'],
    'runits': ['nova-compute/0'],
}


class ODLControllerRelationTest(CharmTestCase):

    def setUp(self):
        super(ODLControllerRelationTest, self).setUp(vpp_data, TO_PATCH)
        self.config.side_effect = self.test_config.get

    def tearDown(self):
        super(ODLControllerRelationTest, self).tearDown()

    @patch.object(charmhelpers.core.hookenv, 'relation_get')
    @patch.object(charmhelpers.core.hookenv, 'related_units')
    @patch.object(charmhelpers.core.hookenv, 'relation_ids')
    def get_odlrel(self, _hrids, _hrunits, _hrget, relinfo=None):
        _hrids.return_value = relinfo['rids']
        _hrunits.return_value = relinfo['runits']
        self.test_relation.set(relinfo['data'])
        _hrget.side_effect = self.test_relation.get
        odl_rel = vpp_data.ODLControllerRelation()
        odl_rel.get_data()
        return odl_rel

    def test_get_first_data(self):
        odl_rel = self.get_odlrel(relinfo=FULL_ODLCTRL)
        expect = {
            'private-address': '10.0.0.27',
            'port': '8080',
            'username': 'odluser',
            'password': 'hardpassword',
        }
        first_data = odl_rel.get_first_data()
        for key in expect.keys():
            self.assertEqual(first_data[key], expect[key])

    def test_get_first_data_missing(self):
        odl_rel = self.get_odlrel(relinfo=EMPTY_ODLCTRL)
        first_data = odl_rel.get_first_data()
        self.assertEqual(first_data, {})

    def test_get_data(self):
        odl_rel = self.get_odlrel(relinfo=FULL_ODLCTRL)
        expect = {
            'odl_ip': '10.0.0.27',
            'odl_port': '8181',
            'odl_username': 'odluser',
            'odl_password': 'hardpassword',
        }
        for key in expect.keys():
            self.assertEqual(odl_rel[key], expect[key])

    def test_is_ready(self):
        odl_rel = self.get_odlrel(relinfo=FULL_ODLCTRL)
        self.assertEqual(odl_rel.is_ready(), True)

    def test_is_ready_incomplete(self):
        odl_rel = self.get_odlrel(relinfo=MISSING_DATA_ODLCTRL)
        self.assertEqual(odl_rel.is_ready(), False)


class NeutronPluginRelationTest(CharmTestCase):

    def setUp(self):
        super(NeutronPluginRelationTest, self).setUp(vpp_data, TO_PATCH)
        self.config.side_effect = self.test_config.get

    def tearDown(self):
        super(NeutronPluginRelationTest, self).tearDown()

    @patch.object(charmhelpers.core.hookenv, 'relation_get')
    @patch.object(charmhelpers.core.hookenv, 'related_units')
    @patch.object(charmhelpers.core.hookenv, 'relation_ids')
    def get_nprel(self, _hrids, _hrunits, _hrget, relinfo=None):
        _hrids.return_value = relinfo['rids']
        _hrunits.return_value = relinfo['runits']
        self.test_relation.set(relinfo['data'])
        _hrget.side_effect = self.test_relation.get
        np_rel = vpp_data.NeutronPluginRelation()
        np_rel.get_data()
        return np_rel

    def test_provide_data(self):
        np_rel = self.get_nprel(relinfo=NP_REL)
        prov_data = np_rel.provide_data()
        self.assertEqual(prov_data['neutron-plugin'], 'odl')
        sub_conf = json.loads(prov_data['subordinate_configuration'])
        nova_conf = sub_conf['nova-compute']['/etc/nova/nova.conf']
        vif_driver = u'nova.virt.libvirt.vif.LibvirtGenericVIFDriver'
        expect = [
            [u'firewall_driver', u'nova.virt.firewall.NoopFirewallDriver'],
            [u'libvirt_vif_driver', vif_driver],
            [u'security_group_api', u'nova'],
        ]
        self.assertEqual(nova_conf['sections']['DEFAULT'], expect)


class ConfigTranslationTest(CharmTestCase):

    def setUp(self):
        super(ConfigTranslationTest, self).setUp(vpp_data, TO_PATCH)
        self.config.side_effect = self.test_config.get

    def tearDown(self):
        super(ConfigTranslationTest, self).tearDown()

    def test_config_default(self):
        ctxt = vpp_data.ConfigTranslation()
        self.assertEqual(ctxt, {'use_syslog': False,
                                'vlan_ranges': 'physnet1:1000:2000',
                                'overlay_network_type': 'vxlan'})

        self.test_config.set('use-syslog', True)
        ctxt = vpp_data.ConfigTranslation()
        self.assertEqual(ctxt, {'use_syslog': True,
                                'vlan_ranges': 'physnet1:1000:2000',
                                'overlay_network_type': 'vxlan'})

        self.test_config.set('vlan-ranges', 'physnet1:1000:3000')
        ctxt = vpp_data.ConfigTranslation()
        self.assertEqual(ctxt, {'use_syslog': True,
                                'vlan_ranges': 'physnet1:1000:3000',
                                'overlay_network_type': 'vxlan'})

    def test_config_exception(self):
        self.test_config.set('overlay-network-type', 'token_ring')
        with nose.tools.assert_raises(ValueError):
            vpp_data.ConfigTranslation()


class PCIInfoTest(CharmTestCase):

    def setUp(self):
        super(PCIInfoTest, self).setUp(vpp_data, TO_PATCH)
        self.config.side_effect = self.test_config.get

    def tearDown(self):
        super(PCIInfoTest, self).tearDown()

    class netdevices():
        def get_macs(self):
            return "fa:16:3e:44:95"

    @patch('vpp_data.os.path.isfile')
    def test_base_match(self, _isfile):
        _isfile.return_value = True
        self.test_config.set(
            'mac-network-map', 'mac=a8:9d:21:cf:93:fd;net=net_d10'
        )
        self.PCINetDevices.side_effect = test_pcidev_helper.pci_devs
        pci_info = vpp_data.PCIInfo()
        self.assertEqual(pci_info['pci_devs'], 'dev 0000:10:00.1')

    @patch('vpp_data.os.path.isfile')
    def test_base_match_but_active(self, _isfile):
        _isfile.return_value = True
        self.test_config.set(
            'mac-network-map', 'mac=a8:9d:21:cf:93:fc;net=net_d10'
        )
        self.PCINetDevices.side_effect = test_pcidev_helper.pci_devs
        pci_info = vpp_data.PCIInfo()
        self.assertEqual(pci_info['pci_devs'], 'no-pci')

    @patch('vpp_data.os.path.isfile')
    def test_base_nomatch(self, _isfile):
        _isfile.return_value = True
        self.test_config.set(
            'mac-network-map', 'mac=d8:dd:d1:df:d3:dd;net=net_d10'
        )
        self.PCINetDevices.side_effect = test_pcidev_helper.pci_devs
        pci_info = vpp_data.PCIInfo()
        self.assertEqual(pci_info['pci_devs'], 'no-pci')

    @patch('vpp_data.os.path.isfile')
    def test_base_multimatch(self, _isfile):
        _isfile.return_value = True
        self.test_config.set(
            'mac-network-map',
            ('mac=a8:9d:21:cf:93:fd;net=net_d10 '
             'mac=84:b8:02:2a:5f:c3;net=net_d20')
        )
        self.PCINetDevices.side_effect = test_pcidev_helper.pci_devs
        pci_info = vpp_data.PCIInfo()
        self.assertEqual(
            pci_info['pci_devs'],
            'dev 0000:10:00.1 dev 0000:06:00.0')

    @patch('vpp_data.os.path.isfile')
    def test_base_multinet(self, _isfile):
        _isfile.return_value = True
        self.test_config.set(
            'mac-network-map',
            ('mac=a8:9d:21:cf:93:fd;net=net_d10 '
             'mac=a8:9d:21:cf:93:fd;net=net_d10')
        )
        self.PCINetDevices.side_effect = test_pcidev_helper.pci_devs
        pci_info = vpp_data.PCIInfo()
        self.assertEqual(pci_info['pci_devs'], 'dev 0000:10:00.1')

    @patch('vpp_data.os.path.isfile')
    def test_base_not_ready(self, _isfile):
        _isfile.return_value = False
        vpp_data.PCIInfo()
        self.log.assert_called_with('PCIInfo not ready')
