from test_utils import CharmTestCase, patch_open
from mock import patch, call
import vpp_utils
import pci_responses

TO_PATCH = [
    'ArchiveUrlFetchHandler',
    'apt_install',
    'apt_purge',
    'charm_dir',
    'config',
    'create_sysctl',
    'fstab_mount',
    'gethostname',
    'hugepage_support',
    'log',
    'rsync',
    'service_restart',
    'unit_private_ip',
    'urllib',
    'user_exists',
]


class _ODLControllerRelation(dict):
    def __init__(self):
        self['odl_ip'] = '10.0.0.10'
        self['odl_port'] = '8181'
        self['odl_username'] = 'admin'
        self['odl_password'] = 'admin'

    def is_ready(self):
        return True


class VPPUtilsTest(CharmTestCase):

    def setUp(self):
        super(VPPUtilsTest, self).setUp(vpp_utils, TO_PATCH)
        self.config.side_effect = self.test_config.get
        self.base_pkgs = [
            'python-psutil', 'gcc', 'make', 'python-crypto', 'qemu', 'seabios'
        ]
        self.charm_dir.return_value = '/var/charmdir'

    def tearDown(self):
        super(VPPUtilsTest, self).tearDown()

    @patch.object(vpp_utils, 'filter_installed_packages')
    def test_install_packages(self, _filter):
        def mocked_filter(pkg_list):
            return pkg_list
        _filter.side_effect = mocked_filter
        vpp_utils.install_packages('servicename')
        self.apt_purge.assert_called_with([])
        self.apt_install.assert_called_with(
            self.base_pkgs + ['corekeeper'])
        self.test_config.set('use-corekeeper', False)
        vpp_utils.install_packages('servicename')
        self.apt_purge.assert_called_with([])
        self.apt_install.assert_called_with(
            self.base_pkgs + ['apport'])

    @patch.object(vpp_utils, 'filter_installed_packages')
    @patch('os.path.isfile')
    @patch('subprocess.check_call')
    @patch('subprocess.check_output')
    def test_install_vpp_packages(self, _check_output, _call, _isfile,
                                  _filter):
        kversion = '3.19.0-21-generic'
        extra_pkg = 'linux-image-extra-{}'.format(kversion)
        deb_url = 'http://10.0.0.10/myurl/vpe.deb'
        _check_output.return_value = kversion
        _filter.return_value = extra_pkg
        filter_calls = [call([extra_pkg]), call(['vpe'])]
        self.test_config.set('vpe-deb-url', deb_url)
        _isfile.return_value = True
        vpp_utils.install_vpp_packages('servicename')
        self.assertEqual(self.config.call_args_list, [call('vpe-deb-url')])
        _filter.assert_has_calls(filter_calls)
        self.urllib.urlretrieve.assert_called_with(
            deb_url, '/var/charmdir/payload/vpe.deb')
        _call.assert_called_with(
            ['dpkg', '-i', '/var/charmdir/payload/vpe.deb'])

    @patch.object(vpp_utils, 'filter_installed_packages')
    @patch('os.path.isfile')
    @patch('subprocess.check_call')
    @patch('subprocess.check_output')
    def test_install_vpp_packages_already_installed(self, _check_output, _call,
                                                    _isfile, _filter):
        kversion = '3.19.0-21-generic'
        _check_output.return_value = kversion
        _filter.return_value = []
        vpp_utils.install_vpp_packages('servicename')
        self.assertEqual(self.config.call_args_list, [])
        self.assertEqual(self.urllib.urlretrieve.call_args_list, [])
        self.assertEqual(self.ArchiveUrlFetchHandler.call_args_list, [])
        self.assertEqual(_call.call_args_list, [])

    @patch.object(vpp_utils, 'filter_installed_packages')
    @patch('os.path.isfile')
    @patch('subprocess.check_call')
    @patch('subprocess.check_output')
    def test_install_vpp_packages_nodeburl(self, _check_output, _call,
                                           _isfile, _filter):
        kversion = '3.19.0-21-generic'
        extra_pkg = 'linux-image-extra-{}'.format(kversion)
        _check_output.return_value = kversion
        _filter.return_value = extra_pkg
        vpp_utils.install_vpp_packages('servicename')
        self.assertEqual(self.urllib.urlretrieve.call_args_list, [])
        self.assertEqual(self.ArchiveUrlFetchHandler.call_args_list, [])
        _call.assert_called_with(
            ['dpkg', '-i', '/var/charmdir/payload/vpe.deb'])

    @patch.object(vpp_utils, 'filter_installed_packages')
    @patch('os.path.isfile')
    @patch('subprocess.check_call')
    @patch('subprocess.check_output')
    def test_install_vpp_packages_deburl_tgz(self, _check_output, _call,
                                             _isfile, _filter):
        kversion = '3.19.0-21-generic'
        extra_pkg = 'linux-image-extra-{}'.format(kversion)
        deb_url = 'http://10.0.0.10/myurl/vpe.deb.tgz'
        _check_output.return_value = kversion
        _filter.return_value = extra_pkg
        self.test_config.set('vpe-deb-url', deb_url)

        class mocked_ArchiveUrlFetchHandler():
            def install(self, url):
                return '/var/charmdir/payload'
        self.ArchiveUrlFetchHandler.side_effect = mocked_ArchiveUrlFetchHandler
        _isfile.return_value = True
        vpp_utils.install_vpp_packages('servicename')
        self.assertEqual(self.urllib.urlretrieve.call_args_list, [])
        self.assertEqual(self.ArchiveUrlFetchHandler.call_args_list, [call()])
        _call.assert_called_with(
            ['dpkg', '-i', '/var/charmdir/payload/vpe.deb'])

    def test_determine_packages(self):
        self.assertEqual(vpp_utils.determine_packages(),
                         self.base_pkgs + ['corekeeper'])
        self.test_config.set('use-corekeeper', False)
        self.assertEqual(vpp_utils.determine_packages(),
                         self.base_pkgs + ['apport'])

    @patch.object(vpp_utils, 'filter_installed_packages')
    @patch.object(vpp_utils, 'bug_1437891')
    def test_purge_packages(self, _bug1437891, _filter):
        def mocked_filter(pkg_list):
            return []
        _filter.side_effect = mocked_filter
        self.test_config.set('use-corekeeper', True)
        vpp_utils.purge_packages()
        self.apt_purge.assert_called_with(['apport'])
        self.test_config.set('use-corekeeper', False)
        vpp_utils.purge_packages()
        _bug1437891.assert_called_with()
        self.apt_purge.assert_called_with(['corekeeper'])

    def test_debug_display_qn_pci_conf(self):
        qnconf_line = ('        args = "unix { nodaemon log /tmp/vpe.log '
                       'cli-listen localhost:5002 full-coredump } api-trace '
                       '{ on } dpdk { socket-mem 1024 dev 0000:00:06.0 }",')
        log_calls = [call('pci config from qn.conf'), call(qnconf_line)]
        with patch_open() as (_open, _file):
            _file.readlines.return_value = pci_responses.QN_CONF.split('\n')
            vpp_utils.debug_display_qn_pci_conf()
        self.log.assert_has_calls(log_calls)

    @patch.object(vpp_utils, 'debug_display_qn_pci_conf')
    @patch.object(vpp_utils, 'check_vpe_cli')
    def test_restart_qn(self, _check_vpe_cli, _debug_display):
        vpp_utils.restart_qn()
        self.log.assert_called_with('Restarting qn')
        _debug_display.assert_called_with()
        self.service_restart.assert_called_with('qn')
        _check_vpe_cli.assert_called_with()

    @patch('subprocess.Popen')
    @patch('subprocess.check_output')
    def test_check_vpe_cli(self, _check_output, _Popen):
        vpp_utils.check_vpe_cli()
        _Popen.assert_called_with(
            [
                'echo',
                '-e',
                'show interfaces-state interface phys-address\nexit'
            ],
            stdout=-1)
        self.assertEqual(
            _check_output.call_args_list[0][0],
            (['/opt/cisco/vpe/bin/confd_cli', '-N', '-C', '-u', 'system'],))

    def test_bug_1437891(self):
        open_calls = [
            call('/var/lib/dpkg/info/corekeeper.postrm', 'r'),
            call('/var/lib/dpkg/info/corekeeper.postrm', 'w'),
        ]
        with patch_open() as (_open, _file):
            _file.readlines.return_value = ['procps reload']
            vpp_utils.bug_1437891()
            _open.assert_has_calls(open_calls)
            self.assertEqual(
                _file.write.call_args_list, [call('procps restart')])

    @patch('psutil.virtual_memory')
    @patch('subprocess.check_call')
    @patch('subprocess.call')
    def test_install_hugepages(self, _call, _check_call, _virt_mem):
        class mem(object):
            def __init__(self):
                self.total = 10000000
        self.user_exists.return_value = True
        self.test_config.set('hugepages', '10%')
        _virt_mem.side_effect = mem
        _call.return_value = 1
        vpp_utils.install_hugepages('servicename')
        self.hugepage_support.assert_called_with(
            'nova',
            mnt_point='/mnt/huge',
            group='root',
            nr_hugepages=488,
            max_map_count=2000,
            mount=False,
        )
        self.create_sysctl.assert_called_with(
            '{kernel.shmmax: 1023410176}\n',
            '/etc/sysctl.d/10-shmmax.conf',
        )
        check_call_calls = [
            call('/etc/init.d/qemu-hugefsdir'),
            call(['update-rc.d', 'qemu-hugefsdir', 'defaults']),
        ]
        _check_call.assert_has_calls(check_call_calls)
        self.fstab_mount.assert_called_with('/mnt/huge')

    @patch('psutil.virtual_memory')
    @patch('subprocess.check_call')
    @patch('subprocess.call')
    def test_install_hugepages_explicit_size(self, _call, _check_call,
                                             _virt_mem):
        self.user_exists.return_value = True
        self.test_config.set('hugepages', '2048')
        vpp_utils.install_hugepages('servicename')
        self.hugepage_support.assert_called_with(
            'nova',
            mnt_point='/mnt/huge',
            group='root',
            nr_hugepages=2048,
            max_map_count=5120,
            mount=False,
        )
        self.create_sysctl.assert_called_with(
            '{kernel.shmmax: 4294967296}\n',
            '/etc/sysctl.d/10-shmmax.conf',
        )

    @patch('psutil.virtual_memory')
    @patch('subprocess.check_call')
    @patch('subprocess.call')
    def test_install_hugepages_already_mounted(self, _call, _check_call,
                                               _virt_mem):
        self.user_exists.return_value = True
        self.test_config.set('hugepages', '2048')
        _call.return_value = 0
        vpp_utils.install_hugepages('servicename')
        self.assertEqual(self.fstab_mount.call_args_list, [])

    @patch('Crypto.PublicKey.RSA.generate')
    @patch('os.chmod')
    @patch('os.path.isfile')
    def test_configure_ssh_keys(self, _isfile, _chmod, _rsa):
        _isfile.return_value = False
        open_calls = [
            call('/opt/cisco/vpe/etc/confd/ssh/ssh_host_rsa_key', 'w'),
            call('/opt/cisco/vpe/etc/confd/ssh/ssh_host_rsa_key.pub', 'w'),
        ]
        with patch_open() as (_open, _file):
            vpp_utils.configure_ssh_keys('servicename')
            _open.assert_has_calls(open_calls)
            _chmod.assert_called_with(
                '/opt/cisco/vpe/etc/confd/ssh/ssh_host_rsa_key', 384)

    @patch('Crypto.PublicKey.RSA.generate')
    @patch('os.path.isfile')
    def test_configure_ssh_keys_existing_key(self, _isfile, _rsa):
        _isfile.return_value = True
        with patch_open() as (_open, _file):
            vpp_utils.configure_ssh_keys('servicename')
            self.assertEqual(_open.call_args_list, [])

    @patch('ODL.ODLConfig')
    @patch('vpp_data.ODLControllerRelation')
    def test_get_odl_config(self, _mock_odlctrl, _odlconfig):
        _mock_odlctrl.side_effect = _ODLControllerRelation
        vpp_utils.get_odl_config()
        _odlconfig.assert_called_with(
            username='admin', ip='10.0.0.10', password='admin', port='8181')

    @patch('ODL.ODLConfig')
    @patch('vpp_data.ODLControllerRelation')
    def test_get_odl_config_notready(self, _mock_odlctrl, _odlconfig):
        class _ODLControllerRelation(dict):
            def is_ready(self):
                return False
        _mock_odlctrl.side_effect = _ODLControllerRelation
        vpp_utils.get_odl_config()
        self.assertEqual(_odlconfig.call_args_list, [])

    @patch('ODL.ODLConfig.odl_register_node')
    @patch('ODL.ODLConfig.is_device_registered')
    @patch('vpp_data.ODLControllerRelation')
    def test_odl_node_registration(self, _mock_odlctrl, _is_registered,
                                   _register):
        self.gethostname.return_value = 'hostname'
        self.unit_private_ip.return_value = '10.0.0.10'
        _mock_odlctrl.side_effect = _ODLControllerRelation
        _is_registered.return_value = False
        vpp_utils.odl_node_registration('servicename')
        _register.assert_called_with('hostname', '10.0.0.10')

    @patch('ODL.ODLConfig.odl_register_node')
    @patch('ODL.ODLConfig.is_device_registered')
    @patch('vpp_data.ODLControllerRelation')
    def test_odl_node_registration_already_registered(self, _mock_odlctrl,
                                                      _is_registered,
                                                      _register):
        self.gethostname.return_value = 'hostname'
        self.unit_private_ip.return_value = '10.0.0.10'
        _mock_odlctrl.side_effect = _ODLControllerRelation
        _is_registered.return_value = True
        vpp_utils.odl_node_registration('servicename')
        self.assertEqual(_register.call_args_list, [])

    @patch('ODL.ODLConfig.odl_register_macs')
    @patch('ODL.ODLConfig.is_net_device_registered')
    @patch('vpp_data.ODLControllerRelation')
    @patch('vpp_data.PCIInfo')
    def test_odl_register_macs(self, _pci_info, _mock_odlctrl,
                               _is_registered, _register):
        _pci_info.return_value = {
            'local_macs': '84:b8:02:2a:5f:c3',
            'local_config': {
                '84:b8:02:2a:5f:c3': [{
                    'net': 'net1',
                    'interface': 'TenGigabitEthernet6/0/0'}
                ],
            }
        }
        self.gethostname.return_value = 'hostname'
        self.unit_private_ip.return_value = '10.0.0.10'
        _mock_odlctrl.side_effect = _ODLControllerRelation
        _is_registered.return_value = False
        vpp_utils.odl_register_macs('servicename')
        _register.assert_called_with(
            'hostname', 'net1', 'TenGigabitEthernet6/0/0', '84:b8:02:2a:5f:c3')

    @patch('ODL.ODLConfig.odl_register_macs')
    @patch('ODL.ODLConfig.is_net_device_registered')
    @patch('vpp_data.ODLControllerRelation')
    @patch('vpp_data.PCIInfo')
    def test_odl_register_macs_already_registered(self, _pci_info,
                                                  _mock_odlctrl,
                                                  _is_registered, _register):
        _pci_info.return_value = {
            'local_macs': ['84:b8:02:2a:5f:c3'],
            'local_config': {
                '84:b8:02:2a:5f:c3': [{
                    'net': 'net1',
                    'interface': 'TenGigabitEthernet6/0/0'}
                ],
            }
        }
        self.gethostname.return_value = 'hostname'
        self.unit_private_ip.return_value = '10.0.0.10'
        _mock_odlctrl.side_effect = _ODLControllerRelation
        _is_registered.return_value = True
        vpp_utils.odl_register_macs('servicename')
        self.assertEqual(_register.call_args_list, [])

    @patch('ODL.ODLConfig.delete_net_device_entry')
    @patch('ODL.ODLConfig.get_macs_networks')
    @patch('vpp_data.ODLControllerRelation')
    @patch('vpp_data.PCIInfo')
    def test_odl_delete_stale_entries_none(self, _pci_info, _mock_odlctrl,
                                           _get_macs_networks, _delete):
        _pci_info.return_value = {
            'local_macs': ['84:b8:02:2a:5f:c3'],
            'local_config': {
                '84:b8:02:2a:5f:c3': [{
                    'net': 'net1',
                    'interface': 'TenGigabitEthernet6/0/0'}
                ],
            }
        }
        self.gethostname.return_value = 'hostname'
        self.unit_private_ip.return_value = '10.0.0.10'
        _mock_odlctrl.side_effect = _ODLControllerRelation
        _get_macs_networks.return_value = ['net1']
        vpp_utils.odl_delete_stale_entries('servicename')
        self.assertEqual(_delete.call_args_list, [])

    @patch('ODL.ODLConfig.delete_net_device_entry')
    @patch('ODL.ODLConfig.get_macs_networks')
    @patch('vpp_data.ODLControllerRelation')
    @patch('vpp_data.PCIInfo')
    def test_odl_delete_stale_entries(self, _pci_info, _mock_odlctrl,
                                      _get_macs_networks, _delete):
        _pci_info.return_value = {
            'local_macs': ['84:b8:02:2a:5f:c3'],
            'local_config': {
                '84:b8:02:2a:5f:c3': [{
                    'net': 'net1',
                    'interface': 'TenGigabitEthernet6/0/0'}
                ],
            }
        }
        self.gethostname.return_value = 'hostname'
        self.unit_private_ip.return_value = '10.0.0.10'
        _mock_odlctrl.side_effect = _ODLControllerRelation
        _get_macs_networks.return_value = ['net2']
        vpp_utils.odl_delete_stale_entries('servicename')
        _delete.assert_called_with('net2', 'hostname')

    @patch('subprocess.check_output')
    def test_filter_installed_packages(self, _check_output):
        _check_output.return_value = pci_responses.DPKG_L
        flump = vpp_utils.filter_installed_packages(['net-tools'])
        self.assertEqual(flump, [])

    @patch('subprocess.check_output')
    def test_filter_installed_packages_not_installed(self, _check_output):
        _check_output.return_value = pci_responses.DPKG_L
        flump = vpp_utils.filter_installed_packages(['mandrake'])
        self.assertEqual(flump, ['mandrake'])

    @patch('vpp_utils.filter_installed_packages')
    def test_vpe_installed(self, _filter):
        _filter.return_value = []
        self.assertTrue(vpp_utils.vpe_installed())

    @patch('vpp_utils.filter_installed_packages')
    def test_vpe_installed_not_installed(self, _filter):
        _filter.return_value = ['vpe']
        self.assertFalse(vpp_utils.vpe_installed())

    @patch('PCIDev.PCINetDevices.__init__')
    @patch('PCIDev.PCINetDevices.rebind_orphans')
    def test_bind_orphaned_net_interfaces(self, _rebind, _pcidevs):
        _pcidevs.return_value = None
        vpp_utils.bind_orphaned_net_interfaces('servicename')
        _rebind.assert_called_with()
